/*!
$Id: cloth.h 642 2009-12-28 05:16:12Z horino $
*/
#ifndef CLOTH_H
#define CLOTH_H

#include "itl/bitset.h"
#include "indi.h"
#include "vector3.h"

namespace indi
{
	namespace physics
	{
		class Particle;
		class Spring;

		/*!
		\brief NXNX
		*/
		class Cloth
		{
		public:
			enum {
				NoStructSpring,
				NoShearSpring,
				NoBendSpring,
			};

			Cloth() : _flags(0), _width(0), _height(0), _nparticle(0), _particles(0), _reset_positions(0), _nspring(0), _springs(0) {}
			Cloth(f32 mass, u32 w, u32 h, u32 f = 0) : _flags(f), _width(w), _height(h), _nparticle(_width * _height), _particles(new Particle * [_nparticle]), _reset_positions(new Vector3 [_nparticle]), _nspring(GetStructSpringNum() + GetShearSpringNum() + GetBendSpringNum()), _springs(new Spring * [_nspring]) {
				Memzero(_particles, sizeof(_particles[0]) * _nparticle);
				Memzero(_springs, sizeof(_springs[0]) * _nspring);
			}
			Cloth(f32 mass, u32 w, u32 h, f32 dx, f32 dy, u32 f = 0) : _flags(f), _width(w), _height(h), _nparticle(_width * _height), _particles(new Particle * [_nparticle]), _reset_positions(new Vector3 [_nparticle]), _nspring(GetStructSpringNum() + GetShearSpringNum() + GetBendSpringNum()), _springs(new Spring * [_nspring]) {
				Memzero(_particles, sizeof(_particles[0]) * _nparticle);
				Memzero(_springs, sizeof(_springs[0]) * _nspring);
				CreateParticles(mass, dx, dy);
				CreateSprings();
			}
			virtual ~Cloth();

			u32 GetStructSpringNum() const { return _flags.Test(NoStructSpring) ? 0 : (_width - 1) * _height + _width * (_height - 1); }
			u32 GetShearSpringNum() const { return _flags.Test(NoShearSpring) ? 0 : ((_width - 1) * (_height - 1) << 1); }
			u32 GetBendSpringNum() const { return _flags.Test(NoBendSpring) ? 0 : (_height * (_width - 2) + _width * (_height - 2)); }

			void Allocate(u32 w, u32 h);
			void CreateParticles(f32 mass, f32 dx, f32 dy);
			void CreateParticles(f32 mass, const Vector3 *positions);
			void CreateSprings(f32 struct_stiffness = 2.0f, f32 struct_dumping = 0.02f, f32 shear_stiffness = 2.0f, f32 shear_dumping = 0.02f, f32 bend_stiffness = 2.0f, f32 bend_dumping = 0.02f);

			u32 GetWidth() const { return _width; }
			u32 GetHeight() const { return _height; }
			const Particle *GetParticle(u32 x, u32 y) const { ASSERT(x * y < _nparticle); return _particles[y * _width + x]; }
			Particle *GetParticle(u32 x, u32 y) { return const_cast<Particle *>(static_cast<const Cloth *>(this)->GetParticle(x, y)); }

			u32 GetParticleNum() const { return _nparticle; }
			const Particle *GetParticle(u32 i) const { ASSERT(i < _nparticle); return _particles[i]; }
			Particle *GetParticle(u32 i) { return const_cast<Particle *>(static_cast<const Cloth *>(this)->GetParticle(i)); }

			u32 GetSpringNum() const { return _nspring; }
			const Spring *GetSpring(u32 i) const { ASSERT(i < _nspring); return _springs[i]; }
			Spring *GetSpring(u32 i) { return const_cast<Spring *>(static_cast<const Cloth *>(this)->GetSpring(i)); }

			void Reset();

		protected:
			itl::Bitset _flags;

			u32 _width;
			u32 _height;

			u32 _nparticle;
			Particle **_particles;
			Vector3 *_reset_positions;

			u32 _nspring;
			Spring **_springs;
		};
	}
}

#endif //< CLOTH_H