/*!
$Id: graphics.h 640 2009-12-24 07:03:53Z horino $
*/
#ifndef GRAPHICS_H
#define GRAPHICS_H

#include <PSGL/psgl.h>
#include <PSGL/psglu.h>

#include "indi.h"
#include "graphics/shader/shader.h"
#include "collision/collision.h"

namespace indi
{
	typedef struct _VIEWPORT _VIEWPORT;
	struct _VIEWPORT
	{
		u32 X;
		u32 Y;  
		u32 Width;
		u32 Height;
		f32 MinZ; 
		f32 MaxZ;
	};

	namespace graphics
	{
		class TextureBase;

		class System
		{
		private:
			System(){}

		public:
			static void Initialize();
			static void Finalize();

			static u32 GetScreenWidth() { return viewport.Width; }
			static u32 GetScreenHeight() { return viewport.Height; }
			static Collision::Frustum &GetFrustum() { return frustum; }
			static const Matrix44 &GetViewportMatrix() { return viewportMatrix; }
			static const Matrix44 &GetInverseViewportMatrix() { return inverseViewportMatrix; }
			static _VIEWPORT GetViewport() { return viewport; }
			static f32 GetAspectRatio() { return aspectRatio; }
			static Vector3 UnProject(const Vector3 &spos, const Matrix44 &world, const Matrix44 &view, const Matrix44 &projection);
			static Vector3 *UnProject(const Vector3 *spos, u32 n, Vector3 *dpos, const Matrix44 &world, const Matrix44 &view, const Matrix44 &projection);
			static Vector3 Project(const Vector3 &wpos, const Matrix44 &wvp_vp);
			static Vector2 ScreenRatio(const Vector3 &center, const Vector3 &radius, const Matrix44 &wvp_vp);

			enum ClearTarget {
#define CLEAR_TARGET(id, gl) id = gl,
#include "graphics.h"
#undef CLEAR_TARGET
			};
			static void ClearColor(u8 r, u8 g, u8 b, u8 a = 255) { glClearColor(static_cast<f32>(r) / 255.0f, static_cast<f32>(g) / 255.0f, static_cast<f32>(b) / 255.0f, static_cast<f32>(a) / 255.0f); }
			static void Clear(u32 m = COLOR | DEPTH | STENCIL) { glClear(m); }

			static void CreateFullScreenShader();
			static void DrawFullScreen(const TextureBase *t);

			static void BeginBackBuffer(){}
			static TextureBase *EndBackBuffer(){ return 0; }
			static void SwapBuffers(graphics::TextureBase *t = 0);
			static void DepthWrite(BOOL b) { glDepthMask(b ? GL_TRUE : GL_FALSE); }
			class AutoBeginBackBuffer
			{
			public:
				AutoBeginBackBuffer() { BeginBackBuffer(); }
				~AutoBeginBackBuffer(){ EndBackBuffer(); }
			};

			enum SampleType {
#define SAMPLE_TYPE(id, val) id = val,
#include "graphics.h"
#undef SAMPLE_TYPE
			};
			static SampleType GetSampleType() { return sampleType; }

			enum FrameRate {
#define FRAME_RATE(id, val) id = val,
#include "graphics.h"
#undef FRAME_RATE
			};
			static FrameRate GetFrameRate() { return frameRate; }

		private:
			static BOOL isVideoReady();

		private:
			static _VIEWPORT viewport;
			static Collision::Frustum frustum;
			static Matrix44 viewportMatrix;
			static Matrix44 inverseViewportMatrix;
			static u32 screenWidth;
			static u32 screenHeight;
			static f32 aspectRatio;
			static SampleType sampleType;
			static FrameRate frameRate;
			class FullScreenPS : public shader::PixelProgram
			{
			public:
				FullScreenPS(const char *path) : shader::PixelProgram(path) {}
				virtual ~FullScreenPS() {}
				void OnLoaded() { shader::PixelProgram::OnLoaded(); sampler = GetParameter("Sampler"); }

				shader::Parameter sampler;
			};
			static shader::VertexProgram *fullscreen_vs;
			static FullScreenPS *fullscreen_ps;

		public:
			class Blend
			{
			public:
				enum BlendTarget {
#define BLEND_TARGET(id, gl) id = gl,
#include "graphics.h"
#undef BLEND_TARGET
				};

				static void Begin() { glEnable(GL_BLEND); }
				static void End() {	glDisable(GL_BLEND); } 
				static void Set(BlendTarget src, BlendTarget dst) {
					glBlendFunc(GL_SRC_ALPHA, src); 
					glBlendFunc(GL_DST_ALPHA, dst); 
				}

				/*!
				\brief XR[v𔲂ƂɎIEnd()R[֗NX
				*/
				class AutoBegin
				{
				public:
					AutoBegin() { Begin(); }
					~AutoBegin(){ End(); }
				private:
				};
			};
		};
	};
};

#endif // GRAPHICS_H

#ifdef CLEAR_TARGET
CLEAR_TARGET(COLOR, GL_COLOR_BUFFER_BIT)
CLEAR_TARGET(DEPTH, GL_DEPTH_BUFFER_BIT)
CLEAR_TARGET(STENCIL, GL_STENCIL_BUFFER_BIT)
#endif

#ifdef BLEND_TARGET
BLEND_TARGET(ZERO, GL_ZERO)
BLEND_TARGET(ONE, GL_ONE)
BLEND_TARGET(SRC_COLOR, GL_SRC_COLOR)
BLEND_TARGET(INV_SRC_COLOR, GL_ONE_MINUS_SRC_COLOR)
BLEND_TARGET(SRC_ALPHA, GL_SRC_ALPHA)
BLEND_TARGET(INV_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
BLEND_TARGET(DST_ALPHA, GL_DST_ALPHA)
BLEND_TARGET(INV_DST_ALPHA, GL_ONE_MINUS_DST_ALPHA)
#endif

#ifdef FRAME_RATE
FRAME_RATE(FRAME_RATE_60, 0)
FRAME_RATE(FRAME_RATE_30, 1)
FRAME_RATE(FRAME_RATE_20, 2)
FRAME_RATE(FRAME_RATE_15, 3)
FRAME_RATE(FRAME_RATE_IMMEDIATE, 4)
#endif

#ifdef SAMPLE_TYPE
SAMPLE_TYPE(SAMPLE_NONE, GL_MULTISAMPLING_NONE_SCE)
SAMPLE_TYPE(SAMPLE_2, GL_MULTISAMPLING_2X_DIAGONAL_CENTERED_SCE)
SAMPLE_TYPE(SAMPLE_4, GL_MULTISAMPLING_4X_SQUARE_ROTATED_SCE/*GL_MULTISAMPLING_4X_SQUARE_CENTERED_SCE*/)
#endif
