/*!
$Id: matrix44.h 645 2010-01-05 09:30:49Z horino $
*/
#ifndef MATRIX44_H
#define MATRIX44_H

#ifdef _DEBUG
#include <iostream>
using namespace std;
#endif

#include <vectormath/cpp/vectormath_soa.h>
using namespace Vectormath;

#include "utilities.h"
#include "indi.h"
#include "vector4.h"

namespace indi
{
	class Vector2;
	class Vector3;
	class Vector4;
	class Quaternion;

	class Matrix44
	{
	private:
		Aos::Matrix4 _e;

	public:
		Matrix44() {}
		Matrix44(const Aos::Matrix4 &v) : _e(v) {}
		Matrix44(f32 v) : _e(Aos::Matrix4(v)) {}
		Matrix44(const f32 *p);

		Matrix44(f32 m00, f32 m01, f32 m02, f32 m03,
			f32 m10, f32 m11, f32 m12, f32 m13,
			f32 m20, f32 m21, f32 m22, f32 m23,
			f32 m30, f32 m31, f32 m32, f32 m33);
		Matrix44(const Quaternion &q);

		f32* toF32(f32 *m) const { 
			m[ 0] = (*this)(0, 0); m[ 1] = (*this)(0, 1); m[ 2] = (*this)(0, 2); m[ 3] = (*this)(0, 3);
			m[ 4] = (*this)(1, 0); m[ 5] = (*this)(1, 1); m[ 6] = (*this)(1, 2); m[ 7] = (*this)(1, 3);
			m[ 8] = (*this)(2, 0); m[ 9] = (*this)(2, 1); m[10] = (*this)(2, 2); m[11] = (*this)(2, 3);
			m[12] = (*this)(3, 0); m[13] = (*this)(3, 1); m[14] = (*this)(3, 2); m[15] = (*this)(3, 3);
			return m;
		}

		u32 Read(const u8 *data) { 
			f32 m[16];
			const u32 *p(reinterpret_cast<const u32 *>(data));
			for(u32 i=0;i<NELEMS(m);++i) 
				m[i] = ToF32(*reinterpret_cast<const u32 *>(p++));
			*this = Matrix44(m);
			return reinterpret_cast<const u8 *>(p) - data;
		}

		operator const Aos::Matrix4&() const { return _e; }
		operator Aos::Matrix4&() { return _e; }
		operator const Aos::Matrix4*() const { return &_e; }
		operator Aos::Matrix4*() { return &_e; }
		operator const Aos::Matrix3() const { return _e.getUpper3x3(); }
		operator Aos::Matrix3() { return _e.getUpper3x3(); }

		Matrix44 &operator=(const Matrix44 &rhs) { _e = rhs._e; return *this; }
		f32 operator[](s32 i) const { ASSERT(i >= 0 && i < 16); return _e[i >> 2][i % 4]; }
		f32 operator()(s32 i, s32 j) const { return _e[i][j]; }
		Matrix44 operator-() const { return Matrix44(-_e); }
		Matrix44 &operator+=(const Matrix44 &rhs) { _e +=rhs._e; return *this; }
		Matrix44 &operator-=(const Matrix44 &rhs) { _e -= rhs._e; return *this; }
		Matrix44 &operator*=(f32 rhs) { _e *= rhs; return *this; }
		Matrix44 &operator*=(const Matrix44 &rhs) { return (*this = *this * rhs); }
		Matrix44 &operator/=(f32 rhs) { _e *= (1.0f / rhs); return *this; }
		f32 Determinant() const { return Aos::determinant(_e); }
		Matrix44 Inverse(f32 /*det*/) const { return Inverse(); }
		Matrix44 Inverse() const { return Matrix44(Aos::inverse(_e)); }
		Matrix44 Transpose() const { return Matrix44(Aos::transpose(_e)); }

		friend BOOL operator==(const Matrix44 &lhs, const Matrix44 &rhs) {
			for(s32 i = 0;i < 4;++i)
				for(s32 j = 0;j < 4;++j)
					if(lhs._e[i][j] != rhs._e[i][j])
						return FALSE;
			return true;
		}
		friend BOOL operator!=(const Matrix44 &lhs, const Matrix44 &rhs) { return !(lhs == rhs); }
		friend Matrix44 operator+(const Matrix44 &lhs, const Matrix44 &rhs){ return Matrix44(lhs._e + rhs._e); }
		friend Matrix44 operator-(const Matrix44 &lhs, const Matrix44 &rhs){ return Matrix44(lhs._e - rhs._e); }
		friend Matrix44 operator*(const Matrix44 &lhs, f32 rhs){ return Matrix44(lhs._e * rhs); }
		friend Matrix44 operator*(f32 lhs, const Matrix44 &rhs){ return Matrix44(lhs * rhs._e); }
		friend Matrix44 operator/(const Matrix44 &lhs, f32 rhs){ return Matrix44(lhs._e * (1.0f / rhs)); }
		friend Matrix44 operator*(const Matrix44 &lhs, const Matrix44 &rhs){ return Matrix44(rhs._e * lhs._e); }
		friend Vector4 operator*(const Matrix44 &lhs, const Vector4 &rhs){ 
			Vector4 v0(lhs[ 0], lhs[ 1], lhs[ 2], lhs[ 3]);
			Vector4 v1(lhs[ 4], lhs[ 5], lhs[ 6], lhs[ 7]);
			Vector4 v2(lhs[ 8], lhs[ 9], lhs[10], lhs[11]);
			Vector4 v3(lhs[12], lhs[13], lhs[14], lhs[15]);
			return Vector4(v0 * rhs, v1 * rhs, v2 * rhs, v3 * rhs);
		}
		friend Vector4 operator*(const Vector4 &lhs, const Matrix44 &rhs){ 
			Vector4 v0(rhs[ 0], rhs[ 4], rhs[ 8], rhs[12]);
			Vector4 v1(rhs[ 1], rhs[ 5], rhs[ 9], rhs[13]);
			Vector4 v2(rhs[ 2], rhs[ 6], rhs[10], rhs[14]);
			Vector4 v3(rhs[ 3], rhs[ 7], rhs[11], rhs[15]);
			return Vector4(v0 * lhs, v1 * lhs, v2 * lhs, v3 * lhs);
		}


		static Matrix44 Zero(){ return Matrix44(0.0f); }
		static Matrix44 Identity(){ return Matrix44(Aos::Matrix4::identity()); }
		static Matrix44 LookAt(const Vector3 &pos, const Vector3 &tag, const Vector3 &up);
		static Matrix44 Perspective(f32 fovy, f32 aspect, f32 zNear, f32 zFar) { return Matrix44(Aos::Matrix4::perspective(fovy, aspect, zNear, zFar)); }
		static Matrix44 Rotation(const Vector3 &v3, f32 rad);
		static Matrix44 RotationX(f32 rad) { return Matrix44(Aos::Matrix4::rotationX(rad)); }
		static Matrix44 RotationY(f32 rad) { return Matrix44(Aos::Matrix4::rotationY(rad)); }
		static Matrix44 RotationZ(f32 rad) { return Matrix44(Aos::Matrix4::rotationZ(rad)); }
		static Matrix44 Translation(const Vector3 &t);
		static Matrix44 Translation(f32 x, f32 y, f32 z);
		static Matrix44 TranslationX(f32 x) { return Translation(x, 0.0f, 0.0f); }
		static Matrix44 TranslationY(f32 y) { return Translation(0.0f, y, 0.0f); }
		static Matrix44 TranslationZ(f32 z) { return Translation(0.0f, 0.0f, z); }
		static Matrix44 Scale(const Vector2 &s);
		static Matrix44 Scale(const Vector3 &s);
		static Matrix44 Scale(f32 x, f32 y, f32 z);
		static Matrix44 Scale(f32 s) { return Scale(s, s, s); }
		static Matrix44 ScaleX(f32 x) { return Scale(x, 1.0f, 1.0f); }
		static Matrix44 ScaleY(f32 y) { return Scale(1.0f, y, 1.0f); }
		static Matrix44 ScaleZ(f32 z) { return Scale(1.0f, 1.0f, z); }

#ifdef _DEBUG
		void NanCheck() const { 
			NAN_CHECK(_e[0][0]); NAN_CHECK(_e[0][1]); NAN_CHECK(_e[0][2]); NAN_CHECK(_e[0][3]);
			NAN_CHECK(_e[1][0]); NAN_CHECK(_e[1][1]); NAN_CHECK(_e[1][2]); NAN_CHECK(_e[1][3]);
			NAN_CHECK(_e[2][0]); NAN_CHECK(_e[2][1]); NAN_CHECK(_e[2][2]); NAN_CHECK(_e[2][3]);
			NAN_CHECK(_e[3][0]); NAN_CHECK(_e[3][1]); NAN_CHECK(_e[3][2]); NAN_CHECK(_e[3][3]);
		}
#endif
	};
};

#ifdef _DEBUG
static ostream ATTR_USED &operator<<(ostream &lhs, const indi::Matrix44 &rhs) { 
	return lhs << "(" << rhs(0, 0) << ", " << rhs(0, 1) << ", " << rhs(0, 2) << ", " << rhs(0, 3) << ")" << endl
		<< "(" << rhs(1, 0) << ", " << rhs(1, 1) << ", " << rhs(1, 2) << ", " << rhs(1, 3) << ")" << endl
		<< "(" << rhs(2, 0) << ", " << rhs(2, 1) << ", " << rhs(2, 2) << ", " << rhs(2, 3) << ")" << endl
		<< "(" << rhs(3, 0) << ", " << rhs(3, 1) << ", " << rhs(3, 2) << ", " << rhs(3, 3) << ")";
}
#endif

#endif //< MATRIX44_H