/*!
$Id: quaternion.h 624 2009-12-17 07:37:52Z horino $
*/
#ifndef QUATERNION_H
#define QUATERNION_H

#ifdef _DEBUG
#include <iostream>
using namespace std;
#endif

#include <vectormath/cpp/vectormath_soa.h>
using namespace Vectormath;

#include "indi.h"
#include "utilities.h"

namespace indi
{
	class Vector3;
	class Vector4;
	//class Matrix33;
	class Matrix44;
	
	class Quaternion
	{
	private:
		Aos::Quat _e;

	public:
		Quaternion(){}
		Quaternion(const Aos::Quat &q) : _e(q) {}
		Quaternion(const Aos::Vector4 &v4) : _e(v4) {}
		Quaternion(f32 v) : _e(Aos::Quat(v)) {}
		Quaternion(f32 x, f32 y, f32 z, f32 w) : _e(Aos::Quat(x, y, z, w)) {}
		Quaternion(const f32 *p) : _e(Aos::Quat(p[0], p[1], p[2], p[3])) {}
		Quaternion(const Vector3 &v3);
		Quaternion(const Vector3 &axis, f32 rad);
		Quaternion(const Vector3 &axisFrom, const Vector3 &axisTo);
		Quaternion(const Vector4 &v4);
		Quaternion(f32 yaw, f32 pitch, f32 roll);
		//Quaternion(const Matrix33 &m33);
		Quaternion(const Matrix44 &m44);

		u32 Read(const u8 *data) { 
			f32 m[4];
			const u32 *p(reinterpret_cast<const u32 *>(data));
			for(u32 i=0;i<NELEMS(m);++i) 
				m[i] = ToF32(*reinterpret_cast<const u32 *>(p++));
			*this = Quaternion(m);
			return reinterpret_cast<const u8 *>(p) - data;
		}

		operator const Aos::Quat&() const { return _e; }
		operator Aos::Quat&() { return _e; }
		operator const Aos::Quat*() const { return &_e; }
		operator Aos::Quat*() { return &_e; }
		
		f32 X() const { return _e.getX(); }
		f32 Y() const { return _e.getY(); }
		f32 Z() const { return _e.getZ(); }
		f32 W() const { return _e.getW(); }
		f32 GetX() const { return X(); }
		f32 GetY() const { return Y(); }
		f32 GetZ() const { return Z(); }
		f32 GetW() const { return W(); }
		Quaternion &Set(f32 x, f32 y, f32 z, f32 w) { _e.setX(x); _e.setY(y); _e.setZ(z); _e.setW(w); return *this; }
		Quaternion &SetX(f32 x) { _e.setX(x); return *this; }
		Quaternion &SetY(f32 y) { _e.setY(y); return *this; }
		Quaternion &SetZ(f32 z) { _e.setZ(z); return *this; }
		Quaternion &SetW(f32 w) { _e.setW(w); return *this; }
		
		Quaternion &operator=(const Quaternion &rhs) { _e = rhs._e; return *this; }
		f32 operator[](s32 i) const { ASSERT(i >= 0 && i < 4); return _e[i]; }
		Quaternion operator~() const { return Conjugate(); }
		Quaternion operator-() const { return Quaternion(-_e); }
		Quaternion &operator+=(const Quaternion &rhs) { _e += rhs._e; return *this; }
		Quaternion &operator-=(const Quaternion &rhs) { _e -= rhs._e; return *this; }
		Quaternion &operator*=(f32 rhs) { _e *=rhs; return *this; }
		Quaternion &operator/=(f32 rhs) { _e /= rhs; return *this; }

		friend Quaternion operator+(const Quaternion &lhs, const Quaternion &rhs){ return Quaternion(lhs) += rhs; }
		friend Quaternion operator-(const Quaternion &lhs, const Quaternion &rhs){ return Quaternion(lhs) -= rhs; }
		friend Quaternion operator*(const Quaternion &lhs, f32 rhs) { return Quaternion(lhs) *= rhs; }
		friend Quaternion operator*(f32 lhs, const Quaternion &rhs) { return Quaternion(rhs) *= lhs; }
		friend Quaternion operator/(const Quaternion &lhs, f32 rhs) { return Quaternion(lhs) /= rhs; }
		friend BOOL operator==(const Quaternion &lhs, const Quaternion &rhs){ return lhs.X() == rhs.X() && lhs.Y() == rhs.Y() && lhs.Z() == rhs.Z() && lhs.W() == rhs.W();	}
		friend BOOL operator!=(const Quaternion &lhs, const Quaternion &rhs){ return !(lhs == rhs); }
		friend Quaternion operator*(const Quaternion &lhs, const Quaternion &rhs){ return Quaternion(lhs._e * rhs._e); }
		friend Quaternion operator*(const Vector3 &lhs, const Quaternion &rhs){ return Quaternion(lhs) * rhs; }
		friend Quaternion operator*(const Quaternion &lhs, const Vector3 &rhs){ return lhs * Quaternion(rhs); }

		f32 Length() const { return Aos::length(_e); }
		f32 LengthSquared() const { return Aos::lengthSqr(Aos::Vector4(_e)); }
		Quaternion &Normalize() { _e = Aos::normalize(_e); return *this; }
		Quaternion Conjugate() const { return Quaternion(Aos::conj(*this)); }
		Quaternion Inverse() const { return (~(*this)).Normalize(); }
		Vector3 Rotate(const Vector3 &v) const;
		void ToAxisRadian(Vector3 &axis, f32 &rad) const;
		
		static Quaternion Identity(){ return Quaternion(Aos::Quat::identity()); }
		static Quaternion RotationX(f32 rad) { return Quaternion(Aos::Quat::rotationX(rad)); }
		static Quaternion RotationY(f32 rad) { return Quaternion(Aos::Quat::rotationY(rad)); }
		static Quaternion RotationZ(f32 rad) { return Quaternion(Aos::Quat::rotationZ(rad)); }
		static Quaternion Lerp(const Quaternion &q0, const Quaternion &q1, f32 t){ return ((1.0f - t) * q0 + t * q1).Normalize(); }
		static Quaternion Slerp(const Quaternion &q0, const Quaternion &q1, f32 t){ return Quaternion(Aos::slerp(t, q0, q1)); }

#ifdef _DEBUG
		void NanCheck() const { NAN_CHECK(_e.getX()); NAN_CHECK(_e.getY()); NAN_CHECK(_e.getZ()); NAN_CHECK(_e.getW()); }
#endif
	};
};

#ifdef _DEBUG
static ostream ATTR_USED &operator<<(ostream &lhs, const indi::Quaternion &rhs) { return lhs << "(" << rhs.X() << ", " << rhs.Y() << ", " << rhs.Z() << ", " << rhs.W() << ")"; }
#endif

#endif //< QUATERNION_H
