/*!
$Id: vector4.h 648 2010-01-07 03:39:59Z horino $
*/
#ifndef VECTOR4_H
#define VECTOR4_H

#ifdef _DEBUG
#include <iostream>
using namespace std;
#endif

#include <vectormath/cpp/vectormath_soa.h>
using namespace Vectormath;	

#include "indi.h"
#include "utilities.h"

namespace indi
{
	class Vector2;
	class Vector3;
	class Quaternion;

	class Vector4
	{
	private:
		Aos::Vector4 _e;

	public:
		Vector4(){}
		Vector4(const Aos::Vector4 &v) : _e(v) {}
		Vector4(f32 v) : _e(Aos::Vector4(v)) {}
		Vector4(f32 x, f32 y, f32 z, f32 w) : _e(Aos::Vector4(x, y, z, w)) {}
		Vector4(const f32 *p) : _e(Aos::Vector4(p[0], p[1], p[2], p[3])) {}
		Vector4(const Vector2 &v2, f32 z, f32 w);
		Vector4(const Vector3 &v3, f32 w);
		Vector4(const Quaternion &q);
		
		u32 Read(const u8 *data) { 
			f32 m[4];
			const u32 *p(reinterpret_cast<const u32 *>(data));
			for(u32 i=0;i<NELEMS(m);++i) 
				m[i] = ToF32(*reinterpret_cast<const u32 *>(p++));
			*this = Vector4(m);			
			return reinterpret_cast<const u8 *>(p) - data;
		}

		operator const Aos::Vector4&() const { return _e; }
		operator Aos::Vector4&() { return _e; }
		operator const Aos::Vector4*() const { return &_e; }
		operator Aos::Vector4*() { return &_e; }

		f32 X() const { return _e.getX(); }
		f32 Y() const { return _e.getY(); }
		f32 Z() const { return _e.getZ(); }
		f32 W() const { return _e.getW(); }
		f32 GetX() const { return X(); }
		f32 GetY() const { return Y(); }
		f32 GetZ() const { return Z(); }
		f32 GetW() const { return W(); }
		Vector4 &Set(f32 x, f32 y, f32 z, f32 w) { _e.setX(x); _e.setY(y); _e.setZ(z); _e.setW(w); return *this; }
		Vector4 &SetX(f32 x) { _e.setX(x); return *this; }
		Vector4 &SetY(f32 y) { _e.setY(y); return *this; }
		Vector4 &SetZ(f32 z) { _e.setZ(z); return *this; }
		Vector4 &SsetW(f32 w) { _e.setW(w); return *this; }

		Vector4 &operator=(const Vector4 &rhs) { _e = rhs._e; return *this; }
		f32 operator[](s32 i) const { ASSERT(i >= 0 && i < 4); return _e[i]; }
		Vector4 operator-() const { return Vector4(-_e); }
		Vector4 &operator+=(const Vector4 &rhs) { _e += rhs._e; return *this; }
		Vector4 &operator-=(const Vector4 &rhs) { _e -= rhs._e; return *this; }
		Vector4 &operator*=(f32 rhs) { _e *= rhs; return *this; }
		Vector4 &operator/=(f32 rhs) { _e /= rhs; return *this; }

		friend Vector4 operator+(const Vector4 &lhs, const Vector4 &rhs){ return Vector4(lhs._e + rhs._e); }
		friend Vector4 operator-(const Vector4 &lhs, const Vector4 &rhs){ return Vector4(lhs._e - rhs._e); }
		friend Vector4 operator*(const Vector4 &lhs, f32 rhs) { return Vector4(lhs._e * rhs); }
		friend Vector4 operator*(f32 lhs, const Vector4 &rhs) { return Vector4(rhs._e * lhs); }
		friend Vector4 operator/(const Vector4 &lhs, f32 rhs) { return Vector4(lhs._e / rhs); }
		friend BOOL operator==(const Vector4 &lhs, const Vector4 &rhs){ return lhs.X() == rhs.X() && lhs.Y() == rhs.Y() && lhs.Z() == rhs.Z() && lhs.W() == rhs.W();	}
		friend BOOL operator!=(const Vector4 &lhs, const Vector4 &rhs){ return !(lhs == rhs); }
		friend f32 operator*(const Vector4 &lhs, const Vector4 &rhs){ return Aos::dot(lhs._e, rhs._e); }

		f32 Length() const { return Aos::length(_e); }
		f32 LengthSquared() const { return Aos::lengthSqr(_e); }
		Vector4 &Normalize() { _e = Aos::normalize(_e); return *this; }
		Vector4 &Adjust(f32 len){ Normalize() *= len; return *this; }

		static f32 Dot(const Vector4 &lhs, const Vector4 &rhs){ return lhs * rhs; }
		static Vector4 Zero(){ return Vector4(0.0f); }
		static Vector4 One(){ return Vector4(1.0f); }
		static Vector4 Half(){ return Vector4(0.5f); }
		static Vector4 UnitX(){ return Vector4(Aos::Vector4::xAxis()); }
		static Vector4 UnitY(){ return Vector4(Aos::Vector4::yAxis()); }
		static Vector4 UnitZ(){ return Vector4(Aos::Vector4::zAxis()); }

#ifdef _DEBUG
		void NanCheck() const { NAN_CHECK(_e.getX()); NAN_CHECK(_e.getY()); NAN_CHECK(_e.getZ()); NAN_CHECK(_e.getW()); }
#endif
	};
};

#ifdef _DEBUG
static ostream ATTR_USED &operator<<(ostream &lhs, const indi::Vector4 &rhs) { return lhs << "(" << rhs.X() << ", " << rhs.Y() << ", " << rhs.Z() << ", " << rhs.W() << ")"; }
#endif

#endif // VECTOR4_H