/*!
$Id: texture.h 635 2009-12-22 08:23:59Z horino $
*/
#ifndef TEXTURE_H
#define TEXTURE_H

#ifdef _DEBUG
#include <iostream>
using namespace std;
#endif

#include "indi.h"
#include "utilities.h"
#include "graphics/graphics.h"
#include "itl/bitset.h"
#include "quaternion_transform.h"

namespace indi
{
	namespace graphics
	{
		/*!
		\brief eNX`x[XNX
		*/
		class TextureBase
		{
		public:
			friend class shader::Parameter;

			enum Filter {
#define TEXFILTER(id, gl) id = gl,
#include "texture.h"
#undef TEXFILTER
			};
			enum Wrap {
#define TEXWRAP(id, gl) id = gl,
#include "texture.h"
#undef TEXWRAP
			};
			enum ColorFormat {
#define COLOR_FORMAT(id, gl) id = gl,
#include "texture.h"
#undef COLOR_FORMAT
			};

			TextureBase(u32 w = 0, u32 h = 0, ColorFormat fmt = TextureFormat()) : _width(w), _height(h), _format(fmt) {}
			TextureBase(ColorFormat fmt) : _width(0), _height(0), _format(fmt) {}
			virtual ~TextureBase() {}

			virtual u32 GetWidth() const { return _width; }
			virtual u32 GetHeight() const { return _height; }
			virtual D3DBaseTexture *GetTexture() { return const_cast<D3DBaseTexture *>(static_cast<const TextureBase *>(this)->GetTexture()); }
			virtual const D3DBaseTexture *GetTexture() const { return const_cast<D3DBaseTexture *>(static_cast<const TextureBase &>(*this).GetTexture()); }

			virtual BOOL IsCubeTexture() const = 0;
			virtual BOOL IsRenderTexture() const = 0;
			virtual BOOL IsAnimationTexture() const = 0;

			virtual u32 Read(const u8 *data);

			ColorFormat Format() { return _format; }

			static ColorFormat TextureFormat() { return ARGB; }
			static ColorFormat RenderTextureFormat() { return ARGB; }
			static ColorFormat DepthTextureFormat() { return SINGLE; }
			static ColorFormat AnimationTextureFormat() { return SINGLE4; }
			//static ColorFormat AnimationTextureFormat() { return HALF4; }

		protected:
			u32 _width;
			u32 _height;
			ColorFormat _format;
		};

		/*!
		\brief eNX`NX
		*/
		class Texture : public TextureBase
		{
		public:
			/*!
			\brief RXgN^
			*/
			Texture() : TextureBase() {}
			Texture(ColorFormat fmt) : TextureBase(fmt) {}
			Texture(u32 w, u32 h, ColorFormat fmt = TextureFormat()) : TextureBase(w, h, fmt), _texture(0) { 
				graphics::System::GetDevice()->CreateTexture(w, h, 1, 0, static_cast<D3DFORMAT>(_format), D3DPOOL_DEFAULT, &_texture, NULL); 
			}			
			Texture(const u8 *data, u32 size) : TextureBase(), _texture(0) { 
				D3DXCreateTextureFromFileInMemory(graphics::System::GetDevice(), data, size, &_texture); 
			}
			virtual ~Texture() { if(_texture) { _texture->Release(); _texture = 0; } }

			u32 GetWidth() const { if(_texture) return _texture->Format.Size.TwoD.Width + 1; return _width; }
			u32 GetHeight() const { if(_texture) return _texture->Format.Size.TwoD.Height + 1; return _height; }

			virtual D3DBaseTexture *GetTexture() { return const_cast<D3DBaseTexture *>(static_cast<const Texture *>(this)->GetTexture()); }
			virtual const D3DBaseTexture *GetTexture() const { return _texture; }

			virtual BOOL IsCubeTexture() const { return FALSE; }
			virtual BOOL IsRenderTexture() const { return FALSE; }
			virtual BOOL IsAnimationTexture() const { return FALSE; }

			u32 Read(const u8 *data);

		protected:
			D3DTexture *_texture;
		};

		/*!
		\brief _OpeNX`NX
		*/
		class RenderTexture : public Texture 
		{
		public:
			RenderTexture(u32 w, u32 h, ColorFormat fmt = RenderTextureFormat(), BOOL create = TRUE) : Texture(w, h, fmt) { 
				if(create)
					graphics::System::GetDevice()->CreateTexture(w, h, 1, D3DUSAGE_RENDERTARGET, static_cast<D3DFORMAT>(fmt), D3DPOOL_DEFAULT, &_texture, NULL); 
			}
			virtual ~RenderTexture() {}

			virtual BOOL IsRenderTexture() const { return TRUE; }
		};

		/*!
		\brief fvX_OpeNX`NX
		*/
		class DepthTexture : public RenderTexture
		{
		public:
			DepthTexture(u32 w, u32 h, ColorFormat fmt = DepthTextureFormat()) : RenderTexture(w, h, fmt) {}
			virtual ~DepthTexture() {}
		};

		/*!
		\brief Aj[VeNX`NX
		*/
		class AnimationTexture : public Texture
		{
		public:
			AnimationTexture(ColorFormat fmt = AnimationTextureFormat()) : Texture(fmt) {}
			virtual ~AnimationTexture() {}

			virtual u32 Read(const u8 *data);

			QuaternionTransform GetTransform(u32 x, u32 y, const u8 *data) const;

			virtual BOOL IsAnimationTexture() const { return TRUE; }
		};

		/*!
		\brief L[ueNX`NX
		*/
		class CubeTexture : public TextureBase
		{
		public:
			/*!
			\brief RXgN^
			*/
			CubeTexture() : TextureBase() {}
			CubeTexture(u32 w, u32 h, ColorFormat fmt = RenderTextureFormat()) : TextureBase(w, h, fmt), _texture(0) {
				graphics::System::GetDevice()->CreateCubeTexture(GetWidth(), 1, 0, static_cast<D3DFORMAT>(_format), D3DPOOL_DEFAULT, &_texture, NULL);
			}
			virtual ~CubeTexture() { if(_texture) { _texture->Release(); _texture = 0; } }

			u32 GetWidth() const { if(_texture) return _texture->Format.Size.TwoD.Width + 1; return _width; }
			u32 GetHeight() const { if(_texture) return _texture->Format.Size.TwoD.Height + 1; return _height; }

			D3DBaseTexture *GetTexture() { return const_cast<D3DBaseTexture *>(static_cast<const CubeTexture *>(this)->GetTexture()); }
			const D3DBaseTexture *GetTexture() const { return _texture; }

			virtual BOOL IsCubeTexture() const { return TRUE; }
			virtual BOOL IsRenderTexture() const { return FALSE; }
			virtual BOOL IsAnimationTexture() const { return FALSE; }

			u32 Read(const u8 *data) {
				u32 size(TextureBase::Read(data));
				D3DXCreateCubeTextureFromFileInMemory(graphics::System::GetDevice(), data, size, &_texture);
				return size;
			}

		private:
			D3DCubeTexture *_texture;
		};
	};
};

#endif // TEXTURE_H

#ifdef TEXFILTER
TEXFILTER(NEAREST, D3DTEXF_POINT)
TEXFILTER(LINEAR, D3DTEXF_LINEAR)
#endif

#ifdef TEXWRAP
TEXWRAP(CLAMP, D3DTADDRESS_CLAMP)
TEXWRAP(REPEAT, D3DTADDRESS_WRAP)
TEXWRAP(MIRROR, D3DTADDRESS_MIRROR)
#endif

#ifdef COLOR_FORMAT
COLOR_FORMAT(ARGB, D3DFMT_A8R8G8B8)
COLOR_FORMAT(SINGLE, D3DFMT_R32F)
COLOR_FORMAT(SINGLE4, D3DFMT_A32B32G32R32F)
COLOR_FORMAT(HALF4, D3DFMT_A16B16G16R16F)
#endif