/*!
$Id: matrix44.h 645 2010-01-05 09:30:49Z horino $
*/
#ifndef MATRIX44_H
#define MATRIX44_H

#ifdef _DEBUG
#include <iostream>
using namespace std;
#endif
#include <D3dx9math.h>

#include "utilities.h"
#include "indi.h"
#include "vector4.h"

namespace indi
{
	class Vector2;
	class Vector3;
	class Vector4;
	class Quaternion;

	class Matrix44
	{
	private:
		D3DXMATRIX _e;

	public:
		Matrix44(){}
		Matrix44(const D3DXMATRIX &m) : _e(m) { }
		Matrix44(const D3DXMATRIX *m) : _e(*m) { }
		Matrix44(f32 v) : _e(v, v, v, v, 
			v, v, v, v, 
			v, v, v, v, 
			v, v, v, v) {}
		Matrix44(f32 *p) : _e(p[0], p[1], p[2], p[3], 
			p[4], p[5], p[6], p[7], 
			p[8], p[9], p[10], p[11], 
			p[12], p[13], p[14], p[15]) {}
		Matrix44(f32 m00, f32 m01, f32 m02, f32 m03,
			f32 m10, f32 m11, f32 m12, f32 m13,
			f32 m20, f32 m21, f32 m22, f32 m23,
			f32 m30, f32 m31, f32 m32, f32 m33) : _e(m00, m01, m02, m03,
			m10, m11, m12, m13,
			m20, m21, m22, m23,
			m30, m31, m32, m33) { }
		Matrix44(const Quaternion &q);
#ifdef _XBOX
		Matrix44(const XMMATRIX &m) : _e(m._11, m._12, m._13, m._14,
			m._21, m._22, m._23, m._24,
			m._31, m._32, m._33, m._34,
			m._41, m._42, m._43, m._44) {}
#endif

		u32 Read(const u8 *data) { 
			f32 m[16];
			const u32 *p(reinterpret_cast<const u32 *>(data));
			for(u32 i=0;i<NELEMS(m);++i) 
				m[i] = ToF32(*reinterpret_cast<const u32 *>(p++));
			*this = Matrix44(m);
			return reinterpret_cast<const u8 *>(p) - data;
		}

		operator const D3DXMATRIX&() const { return _e; }
		operator D3DXMATRIX&() { return _e; }
		operator const D3DXMATRIX*() const { return &_e; }
		operator D3DXMATRIX*() { return &_e; }
		operator const f32*() const { return &_e.m[0][0]; }
		operator f32*() { return &_e.m[0][0]; }
#ifdef _XBOX
		operator XMMATRIX() const {
			return XMMATRIX(_e(0, 0), _e(0, 1), _e(0, 2), _e(0, 3),
				_e(1, 0), _e(1, 1), _e(1, 2), _e(1, 3),
				_e(2, 0), _e(2, 1), _e(2, 2), _e(2, 3),
				_e(3, 0), _e(3, 1), _e(3, 2), _e(3, 3)); 
		}
#endif

		Matrix44 &Set(s32 i, s32 j, f32 v) { _e(i, j) = v; }
		Matrix44 &Set(f32 m00, f32 m01, f32 m02, f32 m03,
			f32 m10, f32 m11, f32 m12, f32 m13,
			f32 m20, f32 m21, f32 m22, f32 m23,
			f32 m30, f32 m31, f32 m32, f32 m33) { 
				_e(0, 0) = m00; _e(0, 1) = m01; _e(0, 2) = m02; _e(0, 3) = m03;
				_e(1, 0) = m10; _e(1, 1) = m11; _e(1, 2) = m12; _e(1, 3) = m13;
				_e(2, 0) = m20; _e(2, 1) = m21; _e(2, 2) = m22; _e(2, 3) = m23;
				_e(3, 0) = m30; _e(3, 1) = m31; _e(3, 2) = m32; _e(3, 3) = m33;
				return *this;
		}

		Matrix44 &operator=(const Matrix44 &rhs) { _e = rhs._e; return *this; }
		f32 operator[](s32 i) const { ASSERT(i >= 0 && i < 16); return _e(i >> 2, i % 4); }
		f32 operator()(s32 i, s32 j) const { return _e(i, j); }
		Matrix44 operator-() const { return Matrix44(-_e); }
		Matrix44 &operator+=(const Matrix44 &rhs) { _e +=rhs._e; return *this; }
		Matrix44 &operator-=(const Matrix44 &rhs) { _e -= rhs._e; return *this; }
		Matrix44 &operator*=(f32 rhs) { _e *= rhs; return *this; }
		Matrix44 &operator*=(const Matrix44 &rhs) { return (*this = *this * rhs); }
		Matrix44 &operator/=(f32 rhs) { _e /= rhs; return *this; }
		f32 Determinant() const { return D3DXMatrixDeterminant(&_e); }
		Matrix44 Inverse(f32 det) const { D3DXMATRIX m; ASSERT(D3DXMatrixInverse(&m, &det, &_e) != NULL); return Matrix44(m); }
		Matrix44 Inverse() const { D3DXMATRIX m; ASSERT(D3DXMatrixInverse(&m, NULL, &_e) != NULL); return Matrix44(m); }
		Matrix44 Transpose() const { D3DXMATRIX m; return Matrix44(D3DXMatrixTranspose(&m, &_e)); }

		friend BOOL operator==(const Matrix44 &lhs, const Matrix44 &rhs){ return lhs._e == rhs._e; }
		friend BOOL operator!=(const Matrix44 &lhs, const Matrix44 &rhs){ return lhs._e != rhs._e; }
		friend Matrix44 operator+(const Matrix44 &lhs, const Matrix44 &rhs){ return Matrix44(lhs._e + rhs._e); }
		friend Matrix44 operator-(const Matrix44 &lhs, const Matrix44 &rhs){ return Matrix44(lhs._e - rhs._e); }
		friend Matrix44 operator*(const Matrix44 &lhs, f32 rhs){ return Matrix44(lhs._e * rhs); }
		friend Matrix44 operator*(f32 lhs, const Matrix44 &rhs){ return Matrix44(lhs * rhs._e); }
		friend Matrix44 operator/(const Matrix44 &lhs, f32 rhs){ return Matrix44(lhs._e / rhs); }
		friend Matrix44 operator*(const Matrix44 &lhs, const Matrix44 &rhs){ return Matrix44(lhs._e * rhs._e); }
		friend Vector4 operator*(const Matrix44 &lhs, const Vector4 &rhs){ 
			Vector4 v0(lhs._e[ 0], lhs._e[ 1], lhs._e[ 2], lhs._e[ 3]);
			Vector4 v1(lhs._e[ 4], lhs._e[ 5], lhs._e[ 6], lhs._e[ 7]);
			Vector4 v2(lhs._e[ 8], lhs._e[ 9], lhs._e[10], lhs._e[11]);
			Vector4 v3(lhs._e[12], lhs._e[13], lhs._e[14], lhs._e[15]);
			return Vector4(v0 * rhs, v1 * rhs, v2 * rhs, v3 * rhs);
		}
		friend Vector4 operator*(const Vector4 &lhs, const Matrix44 &rhs){ 
#if 0
			Vector4 v0(rhs._e[ 0], rhs._e[ 4], rhs._e[ 8], rhs._e[12]);
			Vector4 v1(rhs._e[ 1], rhs._e[ 5], rhs._e[ 9], rhs._e[13]);
			Vector4 v2(rhs._e[ 2], rhs._e[ 6], rhs._e[10], rhs._e[14]);
			Vector4 v3(rhs._e[ 3], rhs._e[ 7], rhs._e[11], rhs._e[15]);
			return Vector4(v0 * lhs, v1 * lhs, v2 * lhs, v3 * lhs);
#else
			D3DXVECTOR4 v; D3DXVec4Transform(&v, lhs, rhs); return Vector4(v);
#endif
		}

		static Matrix44 Zero(){ return Matrix44(0.0f); }
		static Matrix44 Identity(){ D3DXMATRIX m; return Matrix44(*D3DXMatrixIdentity(&m)); }
		static Matrix44 LookAt(const Vector3 &pos, const Vector3 &tag, const Vector3 &up);
		static Matrix44 Perspective(f32 fovy, f32 aspect, f32 zNear, f32 zFar) { D3DXMATRIX m; return Matrix44(*D3DXMatrixPerspectiveFovRH(&m, fovy, aspect, zNear, zFar)); }
		static Matrix44 Rotation(const Vector3 &v3, f32 rad);
		static Matrix44 RotationX(f32 rad) { D3DXMATRIX m; return Matrix44(D3DXMatrixRotationX(&m, rad)); }
		static Matrix44 RotationY(f32 rad) { D3DXMATRIX m; return Matrix44(D3DXMatrixRotationY(&m, rad)); }
		static Matrix44 RotationZ(f32 rad) { D3DXMATRIX m; return Matrix44(D3DXMatrixRotationZ(&m, rad)); }
		static Matrix44 Translation(const Vector3 &t);
		static Matrix44 Translation(f32 x, f32 y, f32 z) { D3DXMATRIX m; return Matrix44(D3DXMatrixTranslation(&m, x, y, z)); }
		static Matrix44 TranslationX(f32 x) { return Translation(x, 0.0f, 0.0f); }
		static Matrix44 TranslationY(f32 y) { return Translation(0.0f, y, 0.0f); }
		static Matrix44 TranslationZ(f32 z) { return Translation(0.0f, 0.0f, z); }
		static Matrix44 Scale(const Vector2 &s);
		static Matrix44 Scale(const Vector3 &s);
		static Matrix44 Scale(f32 x, f32 y, f32 z) { D3DXMATRIX m; return Matrix44(D3DXMatrixScaling(&m, x, y, z)); }
		static Matrix44 Scale(f32 s) { return Scale(s, s, s); }
		static Matrix44 ScaleX(f32 x) { return Scale(x, 1.0f, 1.0f); }
		static Matrix44 ScaleY(f32 y) { return Scale(1.0f, y, 1.0f); }
		static Matrix44 ScaleZ(f32 z) { return Scale(1.0f, 1.0f, z); }

#ifdef _DEBUG
		void NanCheck() const { 
			NAN_CHECK(_e(0, 0)); NAN_CHECK(_e(0, 1)); NAN_CHECK(_e(0, 2)); NAN_CHECK(_e(0, 3));
			NAN_CHECK(_e(1, 0)); NAN_CHECK(_e(1, 1)); NAN_CHECK(_e(1, 2)); NAN_CHECK(_e(1, 3));
			NAN_CHECK(_e(2, 0)); NAN_CHECK(_e(2, 1)); NAN_CHECK(_e(2, 2)); NAN_CHECK(_e(2, 3));
			NAN_CHECK(_e(3, 0)); NAN_CHECK(_e(3, 1)); NAN_CHECK(_e(3, 2)); NAN_CHECK(_e(3, 3));
		}
#endif
	};
};

#ifdef _DEBUG
static ostream &operator<<(ostream &lhs, const indi::Matrix44 &rhs) { 
	return lhs << "(" << rhs(0, 0) << ", " << rhs(0, 1) << ", " << rhs(0, 2) << ", " << rhs(0, 3) << ")" << endl
		<< "(" << rhs(1, 0) << ", " << rhs(1, 1) << ", " << rhs(1, 2) << ", " << rhs(1, 3) << ")" << endl
		<< "(" << rhs(2, 0) << ", " << rhs(2, 1) << ", " << rhs(2, 2) << ", " << rhs(2, 3) << ")" << endl
		<< "(" << rhs(3, 0) << ", " << rhs(3, 1) << ", " << rhs(3, 2) << ", " << rhs(3, 3) << ")";
}
#endif

#endif // MATRIX44_H 