/*!
$Id: quaternion.h 624 2009-12-17 07:37:52Z horino $
*/
#ifndef QUATERNION_H
#define QUATERNION_H

#ifdef _DEBUG
#include <iostream>
using namespace std;
#endif
#include <D3dx9math.h>

#include "indi.h"
#include "utilities.h"

namespace indi
{
	class Vector3;
	class Vector4;
	//class Matrix33;
	class Matrix44;
	
	class Quaternion
	{
	private:
		D3DXQUATERNION _e;

	public:
		Quaternion(){}
		Quaternion(const D3DXQUATERNION &v) : _e(v){}
		Quaternion(const D3DXQUATERNION *v) : _e(*v){}
		Quaternion(f32 v) : _e(v, v, v, v) {}
		Quaternion(f32 x, f32 y, f32 z, f32 w) : _e(x, y, z, w) {}
		Quaternion(const f32 *p) : _e(p[0], p[1], p[2], p[3]) {}
		Quaternion(const Vector3 &v3);
		Quaternion(const Vector3 &axis, f32 rad);
		Quaternion(const Vector3 &from, const Vector3 &to);
		Quaternion(const Vector4 &v4);
		Quaternion(f32 yaw, f32 pitch, f32 roll) { D3DXQuaternionRotationYawPitchRoll(&_e, yaw, pitch, roll); }
		//Quaternion(const Matrix33 &m33);
		Quaternion(const Matrix44 &m44);
#ifdef _XBOX
		Quaternion(const XMVECTOR &v) : _e(v.x, v.y, v.z, v.w) {}
#endif		
		
		u32 Read(const u8 *data) { 
			f32 m[4];
			const u32 *p(reinterpret_cast<const u32 *>(data));
			for(u32 i=0;i<NELEMS(m);++i) 
				m[i] = ToF32(*reinterpret_cast<const u32 *>(p++));
			*this = Quaternion(m);
			return reinterpret_cast<const u8 *>(p) - data;
		}

		operator const D3DXQUATERNION&() const { return _e; }
		operator D3DXQUATERNION&() { return _e; }
		operator const D3DXQUATERNION*() const { return &_e; }
		operator D3DXQUATERNION*() { return &_e; }
		operator const f32*() const { return &_e.x; }
		operator f32*() { return &_e.x; }
#ifdef _XBOX
		operator XMVECTOR() const { return XMVECTOR(*this); }
#endif
		
		f32 X() const { return _e.x; }
		f32 Y() const { return _e.y; }
		f32 Z() const { return _e.z; }
		f32 W() const { return _e.w; }
		f32 GetX() const { return X(); }
		f32 GetY() const { return Y(); }
		f32 GetZ() const { return Z(); }
		f32 GetW() const { return W(); }
		Quaternion &Set(f32 x, f32 y, f32 z, f32 w) { _e.x = x; _e.y = y; _e.z = z; _e.w = w; return *this; }
		Quaternion &SetX(f32 x) { _e.x = x; return *this; }
		Quaternion &SetY(f32 y) { _e.y = y; return *this; }
		Quaternion &SetZ(f32 z) { _e.z = z; return *this; }
		Quaternion &SetW(f32 w) { _e.w = w; return *this; }

		Quaternion &operator=(const Quaternion &rhs) { _e = rhs._e; return *this; }
		f32 operator[](s32 i) { ASSERT(i >= 0 && i < 4); return (&_e.x)[i]; }
		Quaternion operator~() const { return Conjugate(); }
		Quaternion operator-() const { return Quaternion(-_e); }
		Quaternion &operator+=(const Quaternion &rhs) { _e += rhs._e; return *this; }
		Quaternion &operator-=(const Quaternion &rhs) { _e -= rhs._e; return *this; }
		Quaternion &operator*=(f32 rhs) { _e *=rhs; return *this; }
		Quaternion &operator/=(f32 rhs) { _e /= rhs; return *this; }

		friend Quaternion operator+(const Quaternion &lhs, const Quaternion &rhs){ return Quaternion(lhs._e + rhs._e); }
		friend Quaternion operator-(const Quaternion &lhs, const Quaternion &rhs){ return Quaternion(lhs._e - rhs._e); }
		friend Quaternion operator*(const Quaternion &lhs, f32 rhs) { return Quaternion(lhs._e * rhs); }
		friend Quaternion operator*(f32 lhs, const Quaternion &rhs) { return Quaternion(lhs * rhs._e); }
		friend Quaternion operator/(const Quaternion &lhs, f32 rhs) { return Quaternion(lhs._e / rhs); }
		friend BOOL operator==(const Quaternion &lhs, const Quaternion &rhs){ return lhs == rhs; }
		friend BOOL operator!=(const Quaternion &lhs, const Quaternion &rhs){ return lhs != rhs; }
		friend Quaternion operator*(const Quaternion &lhs, const Quaternion &rhs){ D3DXQUATERNION res; return Quaternion(*D3DXQuaternionMultiply(&res, &lhs._e, &rhs._e)); }
		friend Quaternion operator*(const Vector3 &lhs, const Quaternion &rhs){ return Quaternion(lhs) * rhs; }
		friend Quaternion operator*(const Quaternion &lhs, const Vector3 &rhs){ return lhs * Quaternion(rhs); }
		
		f32 Length() const { return D3DXQuaternionLength(&_e); }
		f32 LengthSquared() const { return D3DXQuaternionLengthSq(&_e); }
		Quaternion &Normalize() { D3DXQuaternionNormalize(&_e, &_e); return *this; }
		Quaternion Conjugate() const { D3DXQUATERNION t; return Quaternion(*D3DXQuaternionConjugate(&t, *this)); }
		Quaternion Inverse() const { D3DXQUATERNION t; return Quaternion(*D3DXQuaternionInverse(&t, *this)); }
		Vector3 Rotate(const Vector3 &v) const;
		void ToAxisRadian(Vector3 &axis, f32 &rad) const;
		
		static Quaternion Identity() { D3DXQUATERNION t; return Quaternion(*D3DXQuaternionIdentity(&t)); }
		static Quaternion RotationX(f32 rad);
		static Quaternion RotationY(f32 rad);
		static Quaternion RotationZ(f32 rad);
		static Quaternion Lerp(const Quaternion &q0, const Quaternion &q1, f32 t){ return ((1.0f - t) * q0 + t * q1).Normalize(); }
		static Quaternion Slerp(const Quaternion &q0, const Quaternion &q1, f32 t){ D3DXQUATERNION res; return Quaternion(D3DXQuaternionSlerp(&res, &q0._e, &q1._e, t)); }
#ifdef _DEBUG
		void NanCheck() const { NAN_CHECK(_e.x); NAN_CHECK(_e.y); NAN_CHECK(_e.z); NAN_CHECK(_e.w); }
#endif
	};
};

#ifdef _DEBUG
static ostream &operator<<(ostream &lhs, const indi::Quaternion &rhs) { return lhs << "(" << rhs.X() << ", " << rhs.Y() << ", " << rhs.Z() << ", " << rhs.W() << ")"; }
#endif

#endif // QUATERNION_H