/*!
$Id: vector2.h 648 2010-01-07 03:39:59Z horino $
*/
#ifndef VECTOR2_H
#define VECTOR2_H

#ifdef _DEBUG
#include <iostream>
using namespace std;
#endif
#include <D3dx9math.h>

#include "indi.h"
#include "utilities.h"

namespace indi
{
	class Vector2
	{
	private:
		D3DXVECTOR2 _e;

	public:
		Vector2(){}
		Vector2(const D3DXVECTOR2 &v) : _e(v){}
		Vector2(const D3DXVECTOR2 *v) : _e(*v){}
		Vector2(f32 v) : _e(v, v) {}
		Vector2(f32 x, f32 y) : _e(x, y) {}
		Vector2(const f32 *p) : _e(p[0], p[1]) {}

		u32 Read(const u8 *data) { 
			f32 m[2];
			const u32 *p(reinterpret_cast<const u32 *>(data));
			for(u32 i=0;i<NELEMS(m);++i) 
				m[i] = ToF32(*reinterpret_cast<const u32 *>(p++));
			*this = Vector2(m);
			return reinterpret_cast<const u8 *>(p) - data;
		}

		operator const D3DXVECTOR2&() const { return _e; }
		operator D3DXVECTOR2&() { return _e; }
		operator const D3DXVECTOR2*() const { return &_e; }
		operator D3DXVECTOR2*() { return &_e; }
		operator const f32*() const { return &_e.x; }
		operator f32*() { return &_e.x; }

		f32 X() const { return _e.x; }
		f32 Y() const { return _e.y; }
		f32 GetX() const { return X(); }
		f32 GetY() const { return Y(); }
		Vector2 &Set(f32 x, f32 y) { _e.x = x; _e.y = y; return *this; }
		Vector2 &SetX(f32 x) { _e.x = x; return *this; }
		Vector2 &SetY(f32 y) { _e.y = y; return *this; }

		Vector2 &operator=(const Vector2 &rhs) { _e = rhs._e; return *this; }
		f32 operator[](s32 i) const { ASSERT(i >= 0 && i < 2); return (&_e.x)[i]; }
		Vector2 operator-() const { return Vector2(-_e); }
		Vector2 &operator+=(const Vector2 &rhs) { D3DXVec2Add(&_e, &_e, &rhs._e); return *this; }
		Vector2 &operator-=(const Vector2 &rhs) { D3DXVec2Subtract(&_e, &_e, &rhs._e); return *this; }
		Vector2 &operator*=(f32 rhs) { D3DXVec2Scale(&_e, &_e, rhs); return *this; }
		Vector2 &operator/=(f32 rhs) { _e /= rhs; return *this; }

		friend Vector2 operator+(const Vector2 &lhs, const Vector2 &rhs){ return Vector2(lhs._e + rhs._e); }
		friend Vector2 operator-(const Vector2 &lhs, const Vector2 &rhs){ return Vector2(lhs._e - rhs._e); }
		friend Vector2 operator*(const Vector2 &lhs, f32 rhs) { return Vector2(lhs._e * rhs); }
		friend Vector2 operator*(f32 lhs, const Vector2 &rhs) { return Vector2(lhs * rhs._e); }
		friend Vector2 operator/(const Vector2 &lhs, f32 rhs) { return Vector2(lhs._e / rhs); }
		friend BOOL operator==(const Vector2 &lhs, const Vector2 &rhs){ return lhs._e == rhs._e;	}
		friend BOOL operator!=(const Vector2 &lhs, const Vector2 &rhs){ return lhs._e != rhs._e; }
		friend f32 operator*(const Vector2 &lhs, const Vector2 &rhs){ return D3DXVec2Dot(&lhs._e, &rhs._e); }

		f32 Length() const { return D3DXVec2Length(&_e); }
		f32 LengthSquared() const { return D3DXVec2LengthSq(&_e); }
		Vector2 &Normalize() { D3DXVec2Normalize(&_e, &_e); return *this; }
		Vector2 &Adjust(f32 len){ Normalize() *= len; return *this; }

		static f32 Dot(const Vector2 &lhs, const Vector2 &rhs){ return lhs * rhs; }
		static Vector2 Zero(){ return Vector2(0.0f); }
		static Vector2 One(){ return Vector2(1.0f); }
		static Vector2 Half(){ return Vector2(0.5f); }
		static Vector2 UnitX(){ return Vector2(1.0f, 0.0f); }
		static Vector2 UnitY(){ return Vector2(0.0f, 1.0f); }
#ifdef _DEBUG
		void NanCheck() const { NAN_CHECK(_e.x); NAN_CHECK(_e.y); }
#endif
	};
};

#ifdef _DEBUG
static ostream &operator<<(ostream &lhs, const indi::Vector2 &rhs) { return lhs << "(" << rhs.X() << ", " << rhs.Y() << ")"; }
#endif

#endif // VECTOR2_H