/*!
$Id: vector3.h 648 2010-01-07 03:39:59Z horino $
*/
#ifndef VECTOR3_H
#define VECTOR3_H

#ifdef _DEBUG
#include <iostream>
using namespace std;
#endif
#include <D3dx9math.h>

#include "indi.h"
#include "utilities.h"

namespace indi
{
	class Vector2;
	class Quaternion;
	class Matrix44;
	class Vector4;

	class Vector3
	{
	private:
		D3DXVECTOR3 _e;

	public:
		Vector3(){}
		Vector3(const D3DXVECTOR3 &v) : _e(v){}
		Vector3(const D3DXVECTOR3 *v) : _e(*v){}
		Vector3(f32 v) : _e(v, v, v) {}
		Vector3(f32 x, f32 y, f32 z) : _e(x, y, z) {}
		Vector3(const f32 *p) : _e(p[0], p[1], p[2]) {}
		Vector3(const Vector2 &v2, f32 z);
		Vector3(const Quaternion &q);
		Vector3(const Matrix44 &m44);
		Vector3(const Vector4 &v);
		
		u32 Read(const u8 *data) { 
			f32 m[3];
			const u32 *p(reinterpret_cast<const u32 *>(data));
			for(u32 i=0;i<NELEMS(m);++i) 
				m[i] = ToF32(*reinterpret_cast<const u32 *>(p++));
			*this = Vector3(m);
			return reinterpret_cast<const u8 *>(p) - data;
		}

		operator const D3DXVECTOR3&() const { return _e; }
		operator D3DXVECTOR3&() { return _e; }
		operator const D3DXVECTOR3*() const { return &_e; }
		operator D3DXVECTOR3*() { return &_e; }
		operator const f32*() const { return &_e.x; }
		operator f32*() { return &_e.x; }

		D3DXVECTOR3 *GetD3DVECTOR() { return &_e; }

		f32 X() const { return _e.x; }
		f32 Y() const { return _e.y; }
		f32 Z() const { return _e.z; }
		f32 GetX() const { return X(); }
		f32 GetY() const { return Y(); }
		f32 GetZ() const { return Z(); }
		Vector3 &Set(f32 x, f32 y, f32 z) { _e.x = x; _e.y = y; _e.z = z; return *this; }
		Vector3 &SetX(f32 x) { _e.x = x; return *this; }
		Vector3 &SetY(f32 y) { _e.y = y; return *this; }
		Vector3 &SetZ(f32 z) { _e.z = z; return *this; }

		Vector3 &operator=(const Vector3 &rhs) { _e = rhs._e; return *this; }
		f32 operator[](s32 i) const { ASSERT(i >= 0 && i < 3); return (&_e.x)[i]; }
		Vector3 operator-() const { return Vector3(-_e); }
		Vector3 &operator+=(const Vector3 &rhs) { D3DXVec3Add(&_e, &_e, &rhs._e); return *this; }
		Vector3 &operator-=(const Vector3 &rhs) { D3DXVec3Subtract(&_e, &_e, &rhs._e); return *this; }
		Vector3 &operator*=(f32 rhs) { D3DXVec3Scale(&_e, &_e, rhs); return *this; }
		Vector3 &operator/=(f32 rhs) { _e /= rhs; return *this; }

		friend Vector3 operator+(const Vector3 &lhs, const Vector3 &rhs){ return Vector3(lhs._e + rhs._e); }
		friend Vector3 operator-(const Vector3 &lhs, const Vector3 &rhs){ return Vector3(lhs._e - rhs._e); }
		friend Vector3 operator*(const Vector3 &lhs, f32 rhs) { return Vector3(lhs._e * rhs); }
		friend Vector3 operator*(f32 lhs, const Vector3 &rhs) { return Vector3(lhs * rhs._e); }
		friend Vector3 operator/(const Vector3 &lhs, f32 rhs) { return Vector3(lhs._e / rhs); }
		friend Vector3 operator^(const Vector3 &lhs, const Vector3 &rhs) { D3DXVECTOR3 t; D3DXVec3Cross(&t, &lhs._e, &rhs._e); return Vector3(t); }
		friend BOOL operator==(const Vector3 &lhs, const Vector3 &rhs){ return lhs._e == rhs._e;	}
		friend BOOL operator!=(const Vector3 &lhs, const Vector3 &rhs){ return lhs._e != rhs._e; }
		friend f32 operator*(const Vector3 &lhs, const Vector3 &rhs){ return D3DXVec3Dot(&lhs._e, &rhs._e); }

		f32 Length() const { return D3DXVec3Length(&_e); }
		f32 LengthSquared() const { return D3DXVec3LengthSq(&_e); }
		Vector3 &Normalize() { D3DXVec3Normalize(&_e, &_e); return *this; }
		Vector3 &Adjust(f32 len){ Normalize() *= len; return *this; }

		static f32 Dot(const Vector3 &lhs, const Vector3 &rhs){ return lhs * rhs; }
		static Vector3 Cross(const Vector3 &lhs, const Vector3 &rhs){ return lhs ^ rhs; }
		static Vector3 Zero(){ return Vector3(0.0f); }
		static Vector3 One(){ return Vector3(1.0f); }
		static Vector3 Half(){ return Vector3(0.5f); }
		static Vector3 UnitX(){ return Vector3(1.0f, 0.0f, 0.0f); }
		static Vector3 UnitY(){ return Vector3(0.0f, 1.0f, 0.0f); }
		static Vector3 UnitZ(){ return Vector3(0.0f, 0.0f, 1.0f); }
		static Vector3 Up(){ return UnitY(); }
#ifdef _DEBUG
		void NanCheck() const { NAN_CHECK(_e.x); NAN_CHECK(_e.y); NAN_CHECK(_e.z); }
#endif
	};
};

#ifdef _DEBUG
static ostream &operator<<(ostream &lhs, const indi::Vector3 &rhs) { return lhs << "(" << rhs.X() << ", " << rhs.Y() << ", " << rhs.Z() << ")"; }
#endif

#endif // VECTOR3_H