/*!
$Id: vector4.h 648 2010-01-07 03:39:59Z horino $
*/
#ifndef VECTOR4_H
#define VECTOR4_H

#ifdef _DEBUG
#include <iostream>
using namespace std;
#endif
#include <D3dx9math.h>

#include "indi.h"
#include "utilities.h"

namespace indi
{
	class Vector2;
	class Vector3;
	class Quaternion;

	class Vector4
	{
	private:
		D3DXVECTOR4 _e;

	public:
		Vector4(){}
		Vector4(const D3DXVECTOR4 &v) : _e(v){}
		Vector4(const D3DXVECTOR4 *v) : _e(*v){}
		Vector4(const D3DXPLANE &p) : _e(p.a, p.b, p.c, p.d){}
		Vector4(const D3DXPLANE *p) : _e(p->a, p->b, p->c, p->d){}
		Vector4(f32 v) : _e(v, v, v, v) {}
		Vector4(f32 x, f32 y, f32 z, f32 w) : _e(x, y, z, w) {}
		Vector4(const f32 *p) : _e(p[0], p[1], p[2], p[3]) {}
		Vector4(const Vector2 &v2, f32 z, f32 w);
		Vector4(const Vector3 &v3, f32 w);
		Vector4(const Quaternion &q);
#ifdef _XBOX
		Vector4(const XMVECTOR &v) : _e(v.x, v.y, v.z, v.w) {}
#endif		

		u32 Read(const u8 *data) { 
			f32 m[4];
			const u32 *p(reinterpret_cast<const u32 *>(data));
			for(u32 i=0;i<NELEMS(m);++i) 
				m[i] = ToF32(*reinterpret_cast<const u32 *>(p++));
			*this = Vector4(m);
			return reinterpret_cast<const u8 *>(p) - data;
		}

		operator const D3DXVECTOR4&() const { return _e; }
		operator D3DXVECTOR4&() { return _e; }
		operator const D3DXVECTOR4*() const { return &_e; }
		operator D3DXVECTOR4*() { return &_e; }
		operator const f32*() const { return &_e.x; }
		operator f32*() { return &_e.x; }
#ifdef _XBOX
		operator XMVECTOR() const { return XMVECTOR(*this); }
#endif

		f32 X() const { return _e.x; }
		f32 Y() const { return _e.y; }
		f32 Z() const { return _e.z; }
		f32 W() const { return _e.w; }
		f32 GetX() const { return X(); }
		f32 GetY() const { return Y(); }
		f32 GetZ() const { return Z(); }
		f32 GetW() const { return W(); }
		Vector4 &Set(f32 x, f32 y, f32 z, f32 w) { _e.x = x; _e.y = y; _e.z = z; _e.w = w; return *this; }
		Vector4 &SetX(f32 x) { _e.x = x; return *this; }
		Vector4 &SetY(f32 y) { _e.y = y; return *this; }
		Vector4 &SetZ(f32 z) { _e.z = z; return *this; }
		Vector4 &SetW(f32 w) { _e.w = w; return *this; }

		Vector4 &operator=(const Vector4 &rhs) { _e = rhs._e; return *this; }
		f32 operator[](s32 i) const { ASSERT(i >= 0 && i < 4); return (&_e.x)[i]; }
		Vector4 operator-() const { return Vector4(-_e); }
		Vector4 &operator+=(const Vector4 &rhs) { D3DXVec4Add(&_e, &_e, &rhs._e); return *this; }
		Vector4 &operator-=(const Vector4 &rhs) { D3DXVec4Subtract(&_e, &_e, &rhs._e); return *this; }
		Vector4 &operator*=(f32 rhs) { D3DXVec4Scale(&_e, &_e, rhs); return *this; }
		Vector4 &operator/=(f32 rhs) { _e /= rhs; return *this; }

		friend Vector4 operator+(const Vector4 &lhs, const Vector4 &rhs){ return Vector4(lhs._e + rhs._e); }
		friend Vector4 operator-(const Vector4 &lhs, const Vector4 &rhs){ return Vector4(lhs._e - rhs._e); }
		friend Vector4 operator*(const Vector4 &lhs, f32 rhs) { return Vector4(lhs._e * rhs); }
		friend Vector4 operator*(f32 lhs, const Vector4 &rhs) { return Vector4(lhs * rhs._e); }
		friend Vector4 operator/(const Vector4 &lhs, f32 rhs) { return Vector4(lhs._e / rhs); }
		friend BOOL operator==(const Vector4 &lhs, const Vector4 &rhs){ return lhs._e == rhs._e; }
		friend BOOL operator!=(const Vector4 &lhs, const Vector4 &rhs){ return lhs._e != rhs._e; }
		friend f32 operator*(const Vector4 &lhs, const Vector4 &rhs){ return D3DXVec4Dot(&lhs._e, &rhs._e); }

		f32 Length() const { return D3DXVec4Length(&_e); }
		f32 LengthSquared() const { return D3DXVec4LengthSq(&_e); }
		Vector4 &Normalize() { D3DXVec4Normalize(&_e, &_e); return *this; }
		Vector4 &Adjust(f32 len){ Normalize() *= len; return *this; }

		static f32 Dot(const Vector4 &lhs, const Vector4 &rhs){ return lhs * rhs; }
		static Vector4 Zero(){ return Vector4(0.0f); }
		static Vector4 One(){ return Vector4(1.0f); }
		static Vector4 Half(){ return Vector4(0.5f); }
		static Vector4 UnitX(){ return Vector4(1.0f, 0.0f, 0.0f, 0.0f); }
		static Vector4 UnitY(){ return Vector4(0.0f, 1.0f, 0.0f, 0.0f); }
		static Vector4 UnitZ(){ return Vector4(0.0f, 0.0f, 1.0f, 0.0f); }
#ifdef _DEBUG
		void NanCheck() const { NAN_CHECK(_e.x); NAN_CHECK(_e.y); NAN_CHECK(_e.z); NAN_CHECK(_e.w); }
#endif
	};
};

#ifdef _DEBUG
static ostream &operator<<(ostream &lhs, const indi::Vector4 &rhs) { return lhs << "(" << rhs.X() << ", " << rhs.Y() << ", " << rhs.Z() << ", " << rhs.W() << ")"; }
#endif

#endif // VECTOR4_H